/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * sample of big floats
 */

/* settings for
mpfr_rnd 
0: MPFR_RNDN, round to nearest, with the even rounding rule
              (roundTiesToEven in IEEE 754);
1: MPFR_RNDD, round toward negative infinity
              (roundTowardNegative in IEEE 754)
2: MPFR_RNDU, round toward positive infinity
              (roundTowardPositive in IEEE 754)
3: MPFR_RNDZ, round toward zero
              (roundTowardZero in IEEE 754)
4: MPFR_RNDA, round away from zero
5: MPFR_RNDF, faithful rounding (experimental)

mpfr_precision:
 number of digits after the '.'

mp_random_seed:
specifies the seed used for generating random big floats
*/

/* run */ define test_1() = {
    exists X,Y : 
    {   set mpfr_precision=15 and  
        X=$$2.3333555e+0$$ and Y=$$1.2e+1$$ and stable(Y) and len(10) and
        format("Y=%R\n",Y) and 
        X gets X+Y and always format("X=%15R | X+Y=%15R \n",X,X+Y)
    }
}.

/* run */ define test_2() = {
    exists X : 
    {
        set mpfr_precision=18 and
        X=$$2.3333555$$ and len(10) and X gets sqrt(X) and 
        always format("X=%18R | sqrt(X)=%18R \n",X,sqrt(X))
    }
}.  

/* run */ define test_3() = {
    exists X,Y : 
    {
        set mpfr_precision=15 and
        X=$$2.3333555$$ and Y=1 and stable(Y) and len(10) and
        X gets sqrt(X) + itompfr(Y) and format("Y=%d itompfr(Y)= %R\n",Y,itompfr(Y)) and
        always format("X=%15R | sqrt(X)+itompfr(Y)=%15R \n",X,sqrt(X)+itompfr(Y))
    }
}.

/* run */ define test_4() = {
    exists X,Y : 
    {
        set mpfr_precision=15 and
        X=$$2.3333555$$ and Y=0z1 and stable(Y) and len(10) and
        X gets sqrt(X) + mpztompfr(Y) and format("Y=%Z mpztompfr(Y)= %R\n",Y,mpztompfr(Y)) and
        always format("X=%15R | sqrt(X)+mpztompfr(Y)=%15R \n",X,sqrt(X)+mpztompfr(Y))
    }
}.

/* run */ define test_5() = {
    exists X,Y : 
    {
        set mpfr_precision=15 and
        X=$$1.9333555$$ and Y=1 and stable(Y) and len(10) and
        X gets sqrt(X) + itompfr(Y) and format("Y=%d itompfr(Y)= %R\n",Y,itompfr(Y)) and
        
        always format("X=%15R | sqrt(X)+itompfr(Y)=%15R | ceil(sqrt(X)+itompfr(Y))=%Z \n",X,sqrt(X)+itompfr(Y),ceil(sqrt(X)+itompfr(Y)))
    }
}.

/* run */ define test_6() = {
    exists X,Y : 
    {
        set mpfr_precision=15 and X=$$1.9999999$$ and len(20) and  
        X gets sqrt(X) and 
        always format("X=%15R | sqrt(X)=%15R | floor(sqrt(X))=%Z \n",X,sqrt(X),floor(sqrt(X)))
    }
}.

/* run */ define test_7() = {
    exists X : 
    {
        set mpfr_precision=30 and len(20) and
        set mp_random_seed = 0z0 and
        always format("frrandom=%R \n",frrandom)
    }
}.

/* run */ define test_8() = {
 exists X : {
  list(X,5) and stable(struct X)  and set mpfr_precision=10 and 
   forall i <|X| : { X[i] = frRandom and X[i] gets X[i] + X[i] and len(5)} and
   always output X
 }
}.


/* run */ define test_9() = {
 exists X : {
   set mpfr_precision=10 and empty and
    empty and output ($$1$$/$$0$$) and output ($$0$$/$$0$$) and
    output ($$-1$$/$$0$$)
    
 }
}.

/* run */ define test_10() = {
 exists X : {
   set mpfr_precision=60 and empty and
   format("acos ($$-1$$) = %R\n", acos($$-1$$)) 
    
 }
}.

/* run */ define test_11() = {
 exists X : {
   set mpfr_precision=60 and empty and
   format("exp($$1$$) = %R\n",exp($$1$$) ) 
    
 }
}.

