/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * list operators
 */

define head(L) = {
   L[0]
}.

define tail(L) = {
  L[1..|L|]
}.

define map(f, L) = {
  if L=[] then [] else { [f(head(L))] + map(f, tail(L)) }
}.

define filter(p, L) = {
 if L =[] then []
   else { if p(head(L))
     then [head(L)] + filter(p, tail(L))
       else filter(p, tail(L))
      } 
}.

define zip(L1, L2) = {
 if L1 = [] then []
 else { if L2 = []
        then []
     else [[head(L1),head(L2)]] + zip(tail(L1), tail(L2))
 }
}.

define map_zip(op,L1,L2) = {
 if L1 = [] then []
   else { if L2 = []
          then []
       else [op(head(L1),head(L2))] + map_zip(op,tail(L1),tail(L2))
  }
}.

define foldr(op, L, S) = {
  if L = [] then S else { op(head(L), foldr(op, tail(L), S)) }
}.

define foldl(op, S, L) = {
  if L = [] then S else { foldl(op, op(S, head(L)), tail(L)) }
}.

define fold(op, L, S) = {
  if L = [] then S else { fold(op, tail(L), op(head(L), S)) }
}.

define rev(L) = {
  if L = [] then [] else { rev(tail(L))+[head(L)] } 
}.

define last(L) = {
 if L = [] then "error" else L[|L|-1] 
}.

define butlast(L) = {
 if L = [] then []
   else { if tail(L) = [] then []
       else { [head(L)] + butlast(tail(L)) }
 }
}.

define flatten(L) = {
  if L = [] then [] else { head(L) + flatten(tail(L)) }
}.

define drop(N, L) = {
 if L = [] then []
   else { if N <= 0 then L else drop(N-1, tail(L)) } 
}.

define take(N, L) = {
  if L = [] then []
    else { if N <=0 then [] else [head(L)]+take(N-1,tail(L)) }
}.

define drop_while(p, L) = {
  if L = [] then []
    else { if p(head(L)) then drop_while(p, tail(L)) else L }
}.

define take_while(p, L) = {
  if L = [] then []
    else { if p(head(L)) then [head(L)] + take_while(p, tail(L)) else [] }
}.

define flat_map(f, L) = {
 flatten(map(f, L))
}.



