/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * test samples of list operators
 */

/**************************************************************/
load "list_operators".

define sqr = (lambda x: x*x).

/* run */ define test_1() = {
  exists L,X : {
   L = [i: i<5] and X = map(sqr, L) and
   format("L = %t \n",L) and
   format("sqr = %t \n",sqr) and 
   format("map(sqr, L) = %t\n",X) and empty
 }
}.

define even = (lambda x: x mod 2 = 0).

/* run */ define test_2() = {
  exists L,X : {
    set random_seed=1 and
   L = [Random mod 30 : i< 10] and
   X = filter(even, L) and
   format("L = %t \n",L) and
   format("even = %t \n",even) and 
   format("filter(even, L)  = %t\n",X) and empty
 }
}.

/* run */ define test_3() = {
  exists L1, L2, L : {
   L1 = [i: i<5] and L2 = [i+5 : i<5]  and
   L = zip(L1, L2) and
   format("L1=%t \n",L1) and
   format("L2=%t \n",L2) and
   format("zip(L1, L2) = %t\n",L) and empty
 }
}.

define plus = (lambda x, y : x + y). 

/* (1 + (2 + (3 + (4 + (5 + 6))))) */
/* run */ define test_4() = {
 exists L, X : {
   L = [i+1: i<5] and X = foldr(plus, L, 6) and
   format("L = %t \n",L) and
   format("plus = %t \n",plus) and 
   format("foldr(plus, L, 6) = %t\n",X) and
   empty and output (1 + (2 + (3 + (4 + (5 + 6)))))
 }
}.

/* note this is very slooooooow */
/* run */ define test_4_a() = {
 exists L, X : {
   L = [i+1: i<18] and X = foldr(plus, L, 19) and
   format("L = %t \n",L) and
   format("plus = %t \n",plus) and 
   format("foldr(plus, L, 19) = %t\n",X) and
   empty 
 }
}.

/* (((((0 + 1) + 2) + 3) + 4) + 5) */
/* run */ define test_5() = {
 exists L, X : {
   L = [i+1: i<5] and X = foldl(plus, 0, L) and
   format("L = %t \n",L) and
   format("plus = %t \n",plus) and 
   format("foldl(plus, 0, L) = %t\n",X) and
   empty and output (((((0 + 1) + 2) + 3) + 4) + 5)
 }
}.

/* run */ define test_5_a() = {
 exists L, X : {
   L = [i+1: i<19] and X = foldl(plus, 0, L) and
   format("L = %t \n",L) and
   format("plus = %t \n",plus) and 
   format("foldl(plus, 0, L) = %t\n",X) and
   empty 
 }
}.

/* (5 + (4 + (3 + (2 + (1 + 0))))) */
/* run */ define test_6() = {
 exists L, X : {
   L = [i+1: i<5] and X = fold(plus, L, 0) and
   format("L=%t \n",L) and
   format("plus = %t \n",plus) and 
   format("fold(plus, L, 0) = %t\n",X) and
   empty and output (5 + (4 + (3 + (2 + (1 + 0)))))
 }
}.

/* run */ define test_6_a() = {
 exists L, X : {
   L = [i+1: i<19] and X = fold(plus, L, 0) and
   format("L=%t \n",L) and
   format("plus = %t \n",plus) and 
   format("fold(plus, L, 0) = %t\n",X) and
   empty 
 }
}.

/* run */ define test_7() = {
 exists L, X : {
   L = [i: i<5] and X = rev(L) and
   format("L = %t \n",L) and
   format("rev(L) = %t\n",X) and empty 
 }
}.

/* run */ define test_8() = {
 exists L, X : {
   L = [i: i<6] and X = last(L) and
   format("L = %t \n",L) and
   format("last(L) = %t\n",X) and empty 
 }
}.

/* run */ define test_9() = {
 exists L, X : {
   L = [i: i<6] and X = butlast(L) and
   format("L = %t \n",L) and
   format("butlast(L) = %t\n",X) and empty 
 }
}.

/* run */ define test_10() = {
 exists L, X : {
   L = [ [i+j: j<i] :  i<5] and X = flatten(L) and
   format("L = %t \n",L) and
   format("flatten(L) = %t\n",X) and empty 
 }
}.

/* run */ define test_11() = {
 exists L, X : {
   L = [ i+2 :  i<5] and X = drop(2, L) and
   format("L = %t \n",L) and
   format("drop(2, L) = %t\n",X) and empty 
 }
}.

/* run */ define test_11_a() = {
 exists L, X : {
   L = [ i+2 :  i<5] and X = L[2..|L|] and
   format("L = %t \n",L) and
   format("L[2..|L|] = %t\n",X) and empty 
 }
}.

/* run */ define test_12() = {
 exists L, X : {
   L = [ i+2 :  i<5] and X = take(2, L) and
   format("L = %t \n",L) and
   format("take(2, L) = %t\n",X) and empty 
 }
}.

/* run */ define test_12_a() = {
 exists L, X : {
   L = [ i+2 :  i<5] and X = L[0..2] and
   format("L = %t \n",L) and
   format("L[0..2] = %t\n",X) and empty 
 }
}.


define less_five = (lambda x : x < 5).

/* run */ define test_13() = {
 exists L, X : {
   L = [ i+2 :  i<5] and X = drop_while(less_five, L) and
   format("L = %t \n",L) and
   format("less_five = %t \n",less_five) and
   format("drop_while(less_five, L) = %t\n",X) and empty 
 }
}.

/* run */ define test_14() = {
 exists L, X : {
   L = [ i+2 :  i<5] and X = take_while(less_five, L) and
   format("L = %t \n",L) and
   format("less_five = %t \n",less_five) and
   format("take_while(less_five, L) = %t\n",X) and empty 
 }
}.

/* run */ define test_15() = {
 exists L,X : {
  L = [ [i+j: j<i] :  i<5] and X = flat_map(rev, L) and
  format("L = %t \n",L) and
  format("flat_map(rev, L) = %t\n",X) and empty 
 }
}.

/* run */ define test_16() = {
  exists L,X : {
    L = [ 2*i+1 : i<5 ] and
    X = map_zip( (% A,B: A-B), L[1..|L|], L ) and
    format("L = %t \n",L) and
    format("map_zip( (lambda A,B: A-B), L[1..|L|], L ) = %t \n",X) and empty
  }
}.
