/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: fns.h 284 2017-10-08 08:00:31Z cau $
 */


/* parser.c */
NODEPTR parse(FILE *stream);
NODEPTR parse_string(const char* in);
NODEPTR resolve_refs(NODEPTR e);
/*void push_scope(NODEPTR name_list);*/
/*void push_scope1(NODEPTR name_list, unsigned int history);*/
void push_scope2(NODEPTR name_list, unsigned int history, unsigned framed);
NODEPTR pop_scope(void);
void install_name_list(NODEPTR name_list);
NODEPTR lookup_env(char *s);
NODEPTR install_global_id(char *name);
void install_local_id(char *name);
void syntax_error(int n);
void syntax_msg(int n);


/* lexer.c */
void yyrestart (FILE *input_file  );
void set_input_string(const char* in);
void end_lexical_scan();
NODEPTR symballoc(void);
unsigned hash(char *s);
NODEPTR lookup(char *s);
NODEPTR install(int type, char *name);
void install_keyword(int type, char *name);
void install_reduction(int type, PFI name1, PFI name2, PFI name3);
void install_reduction_w(int type, char *name1, char *name2, char *name3);


/* init.c */
void install_keywords(void);
void initialise_symboltable(void);
void tempura_init (void);


/* top-level.c */
int main(int argc, char **argv);
void prompt(int level);
void tempura(int level);
void exec_command(NODEPTR cmd); 
NODEPTR reduce_stmt(NODEPTR e); 
NODEPTR reduce_val(NODEPTR e); 
NODEPTR reduce_ref(NODEPTR e); 
NODEPTR reduce_pgm(NODEPTR e); 
NODEPTR reduce_ref_and_fetch(NODEPTR e);
NODEPTR reduce_ref_and_fetch1(NODEPTR e);
unsigned int reduce_ref_and_fetch_is_history(NODEPTR e);
unsigned reduce_ref_and_fetch_is_undefined(NODEPTR e);
unsigned reduce_ref_and_fetch_has_default(NODEPTR e);
NODEPTR make_ref(NODEPTR e);
void get_val(NODEPTR e);
void get_ref(NODEPTR e);


/* version.c */
void print_version_number(void);


/* errors.c */
void pop_tempura_level(int b);
void tempura_prefix(char *s);
void error_break1a(char *text, NODEPTR v1);
void error_break2(char *text, char *v1, char *v2);
void error_break2d(char *text, int v1, int v2);
void error_break2e(char *text, int v1);
void user_break(char *text);
void warn1(char *text, char *v1);
void warn2(char *text, char *v1, char *v2);
void show_state(void);
void show_undef_vars(void);
void state_not_defined(void);
void show_form(void);
void show_trace(void);
void show_scope(void);


/* val.c */
void type_check(NODEPTR e, unsigned type);
void static_check(NODEPTR e);
int val_defined(NODEPTR e);
int prev_val_defined(NODEPTR e, unsigned int k);
int curr_val_defined(NODEPTR e);
int next_val_defined(NODEPTR e);
int val_defined2(NODEPTR e);
int pointer_defined(NODEPTR p);
int ilambda_defined(NODEPTR ilambda);
int prev_ilambda_defined(NODEPTR ilambda, unsigned int k);
int curr_ilambda_defined(NODEPTR ilambda);
int next_ilambda_defined(NODEPTR ilambda);     
int list_defined(NODEPTR list);
int prev_list_defined(NODEPTR list, unsigned int k);
int curr_list_defined(NODEPTR list);
int next_list_defined(NODEPTR list);
int max_size(NODEPTR list);
int prev_max_size(NODEPTR list, unsigned int k);
int curr_max_size(NODEPTR list);
int next_max_size(NODEPTR list);
int list_size_defined(NODEPTR list);
int prev_list_size_defined(NODEPTR list, unsigned int k);
int curr_list_size_defined(NODEPTR list);
int next_list_size_defined(NODEPTR list);
void output_val(NODEPTR e);
void output_val_prev(NODEPTR e, unsigned int k);
void output_val_next(NODEPTR e);
void output_val_curr(NODEPTR e);
void foutput_val(NODEPTR e);
void foutput_val_prev(NODEPTR e, unsigned int k);
void foutput_val_next(NODEPTR e);
void foutput_val_curr(NODEPTR e);
void output_ilambda(NODEPTR ilambda);
void output_ilambda_prev(NODEPTR ilambda, unsigned int k);
void output_ilambda_next(NODEPTR ilambda);
void output_ilambda_curr(NODEPTR ilambda);
void foutput_ilambda(NODEPTR ilambda);
void foutput_ilambda_prev(NODEPTR ilambda, unsigned int k);
void foutput_ilambda_next(NODEPTR ilambda);
void foutput_ilambda_curr(NODEPTR ilambda);
void foutput_ilambda(NODEPTR ilambda);
void output_list(NODEPTR list);
void output_list_prev(NODEPTR list, unsigned int k);
void output_list_next(NODEPTR list);
void output_list_curr(NODEPTR list);
void foutput_list(NODEPTR list);
void foutput_list(NODEPTR list);
void foutput_list_prev(NODEPTR list, unsigned int k);
void foutput_list_next(NODEPTR list);
void foutput_list_curr(NODEPTR list);
NODEPTR copy_val(NODEPTR val);
NODEPTR copy_prev_val(NODEPTR val, unsigned int k);
NODEPTR copy_curr_val(NODEPTR val);
NODEPTR copy_next_val(NODEPTR val);
NODEPTR copy_ilambda(NODEPTR ilambda);
NODEPTR copy_curr_ilambda(NODEPTR ilambda);
NODEPTR copy_next_ilambda(NODEPTR ilambda);
NODEPTR copy_prev_ilambda(NODEPTR ilambda, unsigned int k);
NODEPTR copy_list(NODEPTR list);
NODEPTR prev_copy_list(NODEPTR list, unsigned int k);
NODEPTR curr_copy_list(NODEPTR list);
NODEPTR next_copy_list(NODEPTR list);
int string_size(NODEPTR s);
NODEPTR substring(NODEPTR s, int i, int j);
NODEPTR append_string(NODEPTR s1, NODEPTR s2); 
NODEPTR append_list(NODEPTR list1, NODEPTR list2);
/*NODEPTR new_list(int cell_type, int size_type, NODEPTR n);*/
/*NODEPTR new_list1(int cell_type, int size_type, NODEPTR n, unsigned int history);*/
NODEPTR new_list2(int cell_type, int size_type, NODEPTR n, unsigned int history, unsigned framed);
/*void assign_val (register NODEPTR loc, NODEPTR val);*/
void assign_val1 (register NODEPTR loc, NODEPTR val);
void assign_val_update (register NODEPTR loc, NODEPTR val);
void assign_struct(NODEPTR loc, NODEPTR struct1);
void assign_list(NODEPTR loc, NODEPTR list);
void assign_list_update(NODEPTR loc, NODEPTR list);
int equal_val(NODEPTR e1, NODEPTR e2);
int compare_val(NODEPTR e1, NODEPTR e2);
int compare_strings(NODEPTR e1, NODEPTR e2);
int compare_ptrs(NODEPTR e1, NODEPTR e2);
int compare_lists(NODEPTR list1, NODEPTR list2);


/* mem.c */
/*NODEPTR new_loc(register NODEPTR loc, int attrib);*/
/*NODEPTR new_loc1(register NODEPTR loc, int attrib, unsigned int history);*/
NODEPTR new_loc2(register NODEPTR loc, int attrib, unsigned int history, unsigned framed);
/*NODEPTR mem_block(int attrib, register unsigned int size);*/
/*NODEPTR mem_block1(int attrib, register unsigned int size, unsigned int history);*/
NODEPTR mem_block2(int attrib, register unsigned int size, unsigned int history, unsigned framed);
/*NODEPTR allocate_mem(unsigned int size);*/
NODEPTR allocate_mem1(unsigned int size, unsigned int history);
int mem_defined(register NODEPTR loc);
unsigned int mem_defined1(register NODEPTR loc);
NODEPTR get_mem(register NODEPTR loc);
NODEPTR curr_ilambda(NODEPTR ilambda);
NODEPTR next_ilambda(NODEPTR ilambda);
NODEPTR curr_list(NODEPTR list);
NODEPTR pcurr_list(NODEPTR list);
NODEPTR next_list(NODEPTR list);
NODEPTR pnext_list(NODEPTR list);
/*NODEPTR prev_list(NODEPTR list);*/
/*NODEPTR prevnext_list(NODEPTR list);*/
/*NODEPTR prevprev_list(NODEPTR list);*/
NODEPTR list_size1(NODEPTR list);
NODEPTR list_base1(NODEPTR list);
/*NODEPTR put_mem(register NODEPTR loc, NODEPTR val);*/
NODEPTR put_mem1(register NODEPTR loc, NODEPTR val);
int is_static_mem(register NODEPTR loc);
void make_static_mem(register NODEPTR loc);
void make_indirect_mem(NODEPTR loc, register NODEPTR val);
void erase_mem(register NODEPTR loc);
int empty_defined(void);
void put_empty(NODEPTR val);
NODEPTR get_empty(void);
void push_env(NODEPTR bnd);
void pop_env(void);
void push_display(NODEPTR bnd);
void pop_display(void);


/* store.c */
char *allocate(unsigned int size, unsigned int elt_size);
char *strsave(char *s);
char *strapnd(char *s1, char *s2);
char *substr(char *s, int i, int j);
void more_symbol_table(unsigned size);
void more_heap(register unsigned size);
int size_temp_node_list(void);
NODEPTR claim_node(void);
int free_heap_size(void);
NODEPTR node(unsigned type, NODEPTR v1, NODEPTR v2, NODEPTR v3, NODEPTR v4, NODEPTR v5);
NODEPTR bool_node(unsigned int val);
NODEPTR name_node(unsigned signal, char *name, unsigned int level, unsigned int off, unsigned framed);
NODEPTR loc_node(NODEPTR loc, unsigned is_next);
NODEPTR int_node(int64_t val);
NODEPTR string_node(char *val);
NODEPTR float_node(double val);
NODEPTR mpz_node(mpz_t val);
NODEPTR mpz_node_str(char *val);
NODEPTR mpx_node(mpz_t val);
NODEPTR mpx_node_str(char *val);
NODEPTR mpfr_node_str(char *val);
NODEPTR mpfr_node(mpfr_t val);
/*NODEPTR env_node(unsigned level, int size, NODEPTR id);*/
/*NODEPTR env_node1(unsigned level, int size, NODEPTR id, unsigned int history);*/
NODEPTR env_node2(unsigned level, int size, NODEPTR id, unsigned int history, unsigned framed);
NODEPTR bnd_node(NODEPTR env, NODEPTR blk, NODEPTR next);
NODEPTR list_node(NODEPTR blk, NODEPTR size_loc, NODEPTR base_loc);
NODEPTR block_node(unsigned int size, NODEPTR base_loc);
unsigned int elist_size(NODEPTR elist);


/* io.c */
char *titoa(int64_t i);
void output_int(int64_t n, int f);
void foutput_int(int64_t n, int f);
void output_str(char *s);
void foutput_str(char *s);
void output_fl(double n);
void output_mpz(mpz_t n);
void output_mpx(mpz_t n);
void output_mpfr(mpfr_t n);
void foutput_mpfr(mpfr_t n);
void output_mpz_abs(mpz_t n);
void output_mpx_abs(mpz_t n);
void foutput_mpz(mpz_t n);
void foutput_mpx(mpz_t n);
void foutput_mpz_abs(mpz_t n);
void foutput_mpx_abs(mpz_t n);
void foutput_fl(double n);
void outputf_str(char *s, int f);
void foutputf_str(char *s, int f);
void outputf_fl(double n, int f);
void foutputf_fl(double n, int f);
void foutputf_fl_f(double n, int f);
void foutputf_mpz_f(mpz_t n, int f);
void foutputf_mpx_f(mpz_t n, int f);
void foutputf_mpfr_f(mpfr_t n, int f);
void output_lambda(NODEPTR e);
void foutput_lambda(NODEPTR e);
char *print_name(NODEPTR e);
void message(char *text);
void spaces(int n);
void fspaces(int n);
void output_env(NODEPTR env);
void foutput_env(NODEPTR env);
void output_expn(NODEPTR e, int level);
void foutput_expn(NODEPTR e, int level);
void foutput_char(int c);


/* system.c */
void onintr(int sig);
void onfpe(int sig);
void onbus(int sig);
void onsegv(int sig);
void onsys(int sig);
void onalrm(int sig);
void catch_signals(void);
void breakin(void);
FILE *openfile(char *fullname, char *spath, char *name, char *suffix, char *mode);
int fisspace(int s);
void assign_sysvar(NODEPTR id, NODEPTR val);
void init_sysvars(void);
void start_time(void);
void check_time(void);
int timeval_subtract(struct timeval *result, struct timeval *t2, struct timeval *t1);
void timeval_print(struct timeval *tv);
void check_sys_waits(void);


/* r-assign.c */
NODEPTR assign_curr(NODEPTR e1, NODEPTR e2);
NODEPTR assign_next(NODEPTR e1, NODEPTR e2);
void reduce_EQ_stmt(NODEPTR e);
void reduce_LE_stmt(NODEPTR e);
void reduce_LT_stmt(NODEPTR e);
void reduce_GE_stmt(NODEPTR e);
void reduce_GT_stmt(NODEPTR e);
void reduce_NEQ_stmt(NODEPTR e);
void reduce_GETS(NODEPTR e);
void reduce_STB(NODEPTR e);
void reduce_IS(NODEPTR e);
void reduce_DEFAULTS(NODEPTR e);
void reduce_MEM(NODEPTR e);
void reduce_ASSN(NODEPTR e);
void reduce_UASSN(NODEPTR e);
void reduce_SKIPUASSN(NODEPTR e);


/* r-int.c */
void reduce_SEMI(NODEPTR e);
void reduce_iSEMI(NODEPTR e);
void reduce_WHILE(NODEPTR e);
void reduce_REPEAT(NODEPTR e);
void reduce_CHOPSTAR(NODEPTR e);
void reduce_KEEP(NODEPTR e);
void reduce_FORTIMES(NODEPTR e);
void reduce_iFORTIMES(NODEPTR e);
void reduce_FOR(NODEPTR e);
void reduce_iFOR(NODEPTR e);
void reduce_FORIN(NODEPTR e);
void reduce_iFORIN(NODEPTR e);
void reduce_PROJ(NODEPTR e);
void reduce_iPROJ(NODEPTR e);


/* r-linear.c */
void reduce_OR(NODEPTR e);
void reduce_iOR(NODEPTR e);
void reduce_AND(NODEPTR e);
void reduce_iAND(NODEPTR e);
void reduce_IMPLIES(NODEPTR e);
void reduce_WHERE(NODEPTR e);
void reduce_IF_stmt(NODEPTR e);
void reduce_IF_val(NODEPTR e);
void reduce_iIF_val(NODEPTR e);
void reduce_CALL(NODEPTR e);
void reduce_iCALL_stmt(NODEPTR e);
void reduce_iCALL_val(NODEPTR e);
void reduce_iCALL_ref(NODEPTR e);
void reduce_EXISTS(NODEPTR e);
void reduce_iEXISTS_stmt(NODEPTR e);
void reduce_iEXISTS_val(NODEPTR e);
void reduce_FORALL(NODEPTR e);
void reduce_FORALLIN(NODEPTR e);
void reduce_THE(NODEPTR e);
void reduce_ALWAYS(NODEPTR e);
void reduce_SOMETIME(NODEPTR e);
void reduce_HALT(NODEPTR e);
void reduce_INIT(NODEPTR e);
void reduce_FIN(NODEPTR e);
void reduce_KEEP(NODEPTR e);
void reduce_WNEXT(NODEPTR e);
void reduce_NEXT_stmt(NODEPTR e);
void reduce_NEXT_val(NODEPTR e);
void reduce_NEXT_ref(NODEPTR e);
void reduce_PREV_stmt(NODEPTR e);
void reduce_PREV_val(NODEPTR e);
void reduce_PREV_ref(NODEPTR e);
void reduce_NPREV_stmt(NODEPTR e);
void reduce_NPREV_val(NODEPTR e);
void reduce_NPREV_ref(NODEPTR e);
void reduce_LEN(NODEPTR e);
void reduce_iLEN(NODEPTR e);
void reduce_RANLEN(NODEPTR e);
void reduce_INFINITE(NODEPTR e);
void reduce_PROCESS(NODEPTR e);
void reduce_iPROCESS(NODEPTR e);
void reduce_PREFIX(NODEPTR e);
void reduce_iPREFIX(NODEPTR e);


/* r-symbol.c */
void noop(NODEPTR e);
void assert_stmt(NODEPTR e);
void assert_expr(NODEPTR e);
void assert_stmt_to_true(NODEPTR e);
void assert_true(NODEPTR e);
void reduce_const(NODEPTR e);
void reduce_id_val(NODEPTR e);
void reduce_id_ref(NODEPTR e);
void reduce_EMPTY_stmt(NODEPTR e);
void reduce_MORE_stmt(NODEPTR e);
void reduce_EMPTY_val(NODEPTR e);
void reduce_MORE_val(NODEPTR e);
void reduce_EMPTY_ref(NODEPTR e);
void reduce_SKIP(NODEPTR e);


/* r-expr.c */
uint64_t splitmix64(uint64_t x);
void init_xorshift128plus(uint64_t x);
double float_xorshift128plus(void);
uint64_t xorshift128plus(void);
void reduce_TYPE(NODEPTR e);
void reduce_iUNDEFINED(NODEPTR e);
void reduce_EQ_val(NODEPTR e);
int64_t power(int64_t i, int64_t j);
/*int mult32(long a, long b, long* c);*/
int mult(int64_t a, int64_t b, int64_t* c);
NODEPTR mpz_node_plus(NODEPTR e1, NODEPTR e2);
NODEPTR mpfr_node_plus(NODEPTR e1, NODEPTR e2);
NODEPTR mpz_node_minus(NODEPTR e1, NODEPTR e2);
NODEPTR mpfr_node_minus(NODEPTR e1, NODEPTR e2);
NODEPTR mpz_node_mult(NODEPTR e1, NODEPTR e2);
NODEPTR mpfr_node_mult(NODEPTR e1, NODEPTR e2);
NODEPTR mpz_node_div(NODEPTR e1, NODEPTR e2);
NODEPTR mpfr_node_div(NODEPTR e1, NODEPTR e2);
NODEPTR mpz_node_mod(NODEPTR e1, NODEPTR e2);
NODEPTR mpfr_node_mod(NODEPTR e1, NODEPTR e2);
NODEPTR mpz_node_exp_ui(NODEPTR e1, NODEPTR e2);
NODEPTR mpfr_node_exp(NODEPTR e1, NODEPTR e2);
NODEPTR mpz_node_uminus(NODEPTR e1);
NODEPTR mpz_node_itompz(NODEPTR e1);
NODEPTR calculate_val(unsigned op, NODEPTR e1, NODEPTR e2);
void reduce_bin_op(NODEPTR e);
void reduce_u_op(NODEPTR e);
void reduce_STATENUMBER(void);
void reduce_RANDOM(void);
void reduce_RRANDOM(void);
void reduce_SRANDOM(void);
void reduce_SRRANDOM(void);
void reduce_ZRANDOM(void);
void reduce_ZRRANDOM(void);
void reduce_FRANDOM(void);
void reduce_FRRANDOM(void);
void reduce_ASCII(NODEPTR e);
void reduce_REFER(NODEPTR e);
void reduce_DEREF(NODEPTR e);


/* r-io.c */
void reduce_WINPUT(NODEPTR e);
void reduce_INPUT(NODEPTR e);
void reduce_OUTPUT(NODEPTR e);
void reduce_iREMEMBER(NODEPTR e);
void reduce_PEEK(NODEPTR e);
void reduce_FORMAT(NODEPTR e);


/* r-lists.c */
void reduce_list(NODEPTR e);
void reduce_STRUCT_val(NODEPTR e);
void reduce_STRUCT_ref(NODEPTR e);
void reduce_MAKELIST(NODEPTR e);
void reduce_MAKERANGELIST(NODEPTR e);
int reduce_index(NODEPTR e);
void reduce_SUB(NODEPTR e);
void reduce_SIZE(NODEPTR e);
NODEPTR sublist(NODEPTR list, int m, int n);
void reduce_SUBLIST(NODEPTR e);


/* r-system.c */
void reduce_RUN(NODEPTR e);
void init_exec_stat(void);
void init_state_stat(void);
void init_pass_stat(void);
void update_exec_stat(void);
void update_state_stat(void);
void output_pass_stat(void);
void output_state_stat(void);
void output_exec_stat(void);
void reduce_LOAD(NODEPTR e);
void reduce_DEFINE(NODEPTR e);
void reduce_EXIT(NODEPTR e);
void reduce_SET(NODEPTR e);
void reduce_SYSVAR(NODEPTR e);
void reduce_SHOWTRACE(NODEPTR e);
/*void show_mem(NODEPTR loc);*/
void show_mem1(NODEPTR loc);
/*void show_mem_all(NODEPTR loc);*/
void show_mem_all1(NODEPTR loc);
void show_mem_nodef(NODEPTR loc);
void reduce_SHOWMEM(NODEPTR e);
void reduce_EOF(NODEPTR e);
void reduce_RETURN(NODEPTR e);
void reduce_CONTINUE(NODEPTR e);
void reduce_LEAP(NODEPTR e);
void reduce_STEP(NODEPTR e);
void reduce_BREAK(NODEPTR e);
void reduce_ABORT(NODEPTR e);
void reduce_HELP(NODEPTR e);
void reduce_SYSTEM(NODEPTR e);
void reduce_iSYSTEM(NODEPTR e);
void reduce_SHOWSTATE(NODEPTR e);


/* gc.c */
void mark(NODEPTR e);
void mark_mem(NODEPTR loc, unsigned n);
void mark_phase(void);
int is_refd_block(NODEPTR bp);
void reclaim_node(NODEPTR np);
void reclaim_phase(void);
void collect_garbage(void);




