/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-expr.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"

/* The following are from http://xorshift.di.unimi.it/ */
/* slightly adapted for Tempura                        */
uint64_t splitmix64(uint64_t x)
{
	uint64_t z = (x += UINT64_C(0x9E3779B97F4A7C15));
	z = (z ^ (z >> 30)) * UINT64_C(0xBF58476D1CE4E5B9);
	z = (z ^ (z >> 27)) * UINT64_C(0x94D049BB133111EB);
	return z ^ (z >> 31);
}

uint64_t s[2];

void init_xorshift128plus(uint64_t x)
{
    struct timeval tnow;
    uint64_t seed64;
    if (x == 0) {
        gettimeofday(&tnow, NULL);
        seed64 = (uint64_t)tnow.tv_usec;
        if ( seed64 == 0 ) {
            seed64 = UINT64_C(42);
        }
    } else {
        seed64 = x;
    }
    printf("Using random_seed = %" PRId64 "\n",seed64);
    s[0] = splitmix64(seed64);
    s[1] = splitmix64(s[0]);
}

double float_xorshift128plus(void) 
{
    uint64_t x = xorshift128plus();
    double d;
    
    x = UINT64_C(0x3FF) << 52 | x >> 12;
    d = *((double *)&x) - 1.0;

    return d;
}


uint64_t xorshift128plus(void)
{
        if (s[0] == 0 && s[1] == 0) {
            init_xorshift128plus(0);
        }
        uint64_t s1 = s[0];
	const uint64_t s0 = s[1];
	s[0] = s0;
	s1 ^= s1 << 23; // a
	s[1] = s1 ^ s0 ^ (s1 >> 18) ^ (s0 >> 5); // b, c
	return s[1] + s0; 
}

/* end of xorshift */

/****************************************\
*                                        *
*  statements and expressions for types  *
*                                        *
\****************************************/


/* Get the type of something. */
/* type e */
void reduce_TYPE(NODEPTR e)
{
    NODEPTR loc =reduce_ref  (node_lnk_1(e));

    if (Ready) {
       if ((Ready = mem_defined1 (loc)))
           result (string_node (strsave (print_name (get_mem (loc)))));
    }
}


/* Check whether e is undefined. */
/* _undefined e */
void reduce_iUNDEFINED(NODEPTR e)
{
     NODEPTR loc=reduce_ref (node_lnk_1(e));
     
     if (DebugLevel>=5) {
         printf("<< FO=%1d, SPF=%1d, PPF=%1d, Rdy=%1d, Rtry=%1d>>\n",ForceOutput,StateProgressFlag,PassProgressFlag,Ready,Retry);
     }
         /*if ((Ready && mem_defined1 (loc)) || ForceOutput) {*/
     if ((Ready &&  mem_defined1 (loc))|| ( ForceOutput)) {
             if (mem_defined1 (loc)){
                 if (val_defined2(get_mem(loc))) {
                     node_mem_isundefined(loc)=0;
                     result(False);
                 }
                 else {
                   node_mem_isundefined(loc)=1;
                   result(True);  
                 }
             }
             else{
                 node_mem_isundefined(loc)=1;
                 result(True);
             }
         } 
             /* }*/
     
    
         Ready = (Ready &&  mem_defined1 (loc)) || (ForceOutput) ;
}

/**************************************\
*                                      *
*  Simple Boolean, arithmetic, string  *
*  and list expressions.               *
*                                      *
\**************************************/

/*Equality and Non-equality. */
/* e1 = e2, e1 ~= e2 */
void reduce_EQ_val(NODEPTR e)
{
    NODEPTR e1, e2;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        e2 = reduce_val (node_lnk_2(e));
        if (Ready) {
            if (node_type(e) == EQ)
                result (equal_val (e1, e2) ? True : False);
            else
                result (equal_val (e1, e2) ? False : True);
        }
    }
}


#define __HALF_MAX_SIGNED(type) ((type)1 << (sizeof(type)*8-2))
#define __MAX_SIGNED(type) (__HALF_MAX_SIGNED(type) - 1 + __HALF_MAX_SIGNED(type))
#define __MIN_SIGNED(type) (-1 - __MAX_SIGNED(type))

#define __MIN(type) ((type)-1 < 1?__MIN_SIGNED(type):(type)0)
#define __MAX(type) ((type)~__MIN(type))

#define assigncheck(dest,src) ({ \
  typeof(src) __x=(src); \
  typeof(dest) __y=__x; \
  (__x==__y && ((__x<1) == (__y<1))?(void)((dest)=__y),0:1); \
})

#define add_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MIN(typeof(c))-(__b)<=(__a))?assigncheck(c,__a+__b):1) : \
    ((__MAX(typeof(c))-(__b)>=(__a))?assigncheck(c,__a+__b):1); \
})

#define sub_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MAX(typeof(c))+(__b)>=(__a))?assigncheck(c,__a-__b):1) : \
    ((__MIN(typeof(c))+(__b)<=(__a))?assigncheck(c,__a-__b):1); \
})


/* buggy one
#define sub_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MAX(typeof(c))-(__b)>=(__a))?assigncheck(c,__a-__b):1) : \
                      ^ oops
    ((__MIN(typeof(c))+(__b)<=(__a))?assigncheck(c,__a-__b):1); \
})
*/

/*int mult32(long a,long b,long* c) {
    long long x=(long long)a*b;
    if (x!=(long)x) return 0;
  *c=x;
  return 1;
  }*/


/*#if defined(__LP64__)*/
int mult(int64_t x, int64_t y, int64_t * result)
{
    *result = 0;
    if (x > 0 && y > 0 && x > INT64_MAX / y) return 1;
    if (x < 0 && y > 0 && x < INT64_MIN / y) return 1;
    if (x > 0 && y < 0 && y < INT64_MIN / x) return 1;
    if (x < 0 && y < 0 && (x <= INT64_MIN || y <= INT64_MIN || -x > INT64_MAX / -y))
        return 1;
    *result = x * y;
    return 0;
}

/*#else
int mult(long x, long y, long * result)
{
    *result = 0;
    if (x > 0 && y > 0 && x > INT32_MAX / y) return 1;
    if (x < 0 && y > 0 && x < INT32_MIN / y) return 1;
    if (x > 0 && y < 0 && y < INT32_MIN / x) return 1;
    if (x < 0 && y < 0 && (x <= INT32_MIN || y <= INT32_MIN || -x > INT32_MAX / -y))
        return 1;
    *result = x * y;
    return 0;
}
#endif
*/

int64_t power(int64_t i, int64_t j)
{
    int64_t p = 1;
    int64_t product;
    while (j-- > 0){
        assert0(mult (p, i, &product) == 0,"integer overflow in power" );
        p = product;
    }
    return (p);
}

#define int_plus(X,Y) int_node (node_ival(X) + node_ival(Y))
#define float_plus(X,Y) float_node(node_flval(X) + node_flval(Y))

NODEPTR mpz_node_plus(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpz_add(node_mpzval(fp), node_mpzval(e1),node_mpzval(e2));
  return (fp);
}

NODEPTR mpx_node_plus(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_add(node_mpxval(fp), node_mpxval(e1),node_mpxval(e2));
  return (fp);
}


NODEPTR mpfr_node_plus(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_add(node_mpfrval(fp), node_mpfrval(e1),node_mpfrval(e2), MPFRRnd);
  return (fp);
}

#define int_minus(X,Y) int_node (node_ival(X) - node_ival(Y))
#define float_minus(X,Y) float_node(node_flval(X) - node_flval(Y))

NODEPTR mpz_node_minus(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpz_sub(node_mpzval(fp), node_mpzval(e1),node_mpzval(e2));
  return (fp);
}

NODEPTR mpx_node_minus(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_sub(node_mpxval(fp), node_mpxval(e1),node_mpxval(e2));
  return (fp);
}

NODEPTR mpfr_node_minus(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_sub(node_mpfrval(fp), node_mpfrval(e1),node_mpfrval(e2), MPFRRnd);
  return (fp);
}

#define int_mult(X,Y) int_node (node_ival(X) * node_ival(Y))
#define float_mult(X,Y) float_node(node_flval(X) * node_flval(Y))

NODEPTR mpz_node_mult(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpz_mul(node_mpzval(fp), node_mpzval(e1),node_mpzval(e2));
  return (fp);
}

NODEPTR mpx_node_mult(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_mul(node_mpxval(fp), node_mpxval(e1),node_mpxval(e2));
  return (fp);
}

NODEPTR mpfr_node_mult(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_mul(node_mpfrval(fp), node_mpfrval(e1),node_mpfrval(e2), MPFRRnd);
  return (fp);
}

#define int_div(X,Y) int_node (node_ival(X) / node_ival(Y))
#define float_div(X,Y) float_node(node_flval(X) / node_flval(Y))

NODEPTR mpz_node_div(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpz_cdiv_q(node_mpzval(fp), node_mpzval(e1),node_mpzval(e2));
  return (fp);
}

NODEPTR mpx_node_div(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_cdiv_q(node_mpxval(fp), node_mpxval(e1),node_mpxval(e2));
  return (fp);
}

NODEPTR mpfr_node_div(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_div(node_mpfrval(fp), node_mpfrval(e1),node_mpfrval(e2), MPFRRnd);
  return (fp);
}

#define int_mod(X,Y) int_node (mod (node_ival(X), node_ival(Y)))
#define float_mod(X,Y) float_node(fmod(node_flval(X),node_flval(Y)))

NODEPTR mpz_node_mod(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpz_mod(node_mpzval(fp), node_mpzval(e1),node_mpzval(e2));
  return (fp);
}

NODEPTR mpx_node_mod(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_mod(node_mpxval(fp), node_mpxval(e1),node_mpxval(e2));
  return (fp);
}

NODEPTR mpfr_node_mod(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_fmod(node_mpfrval(fp), node_mpfrval(e1),node_mpfrval(e2), MPFRRnd);
  return (fp);
}

#define float_atantwo(X,Y) float_node(atan2(node_flval(X),node_flval(Y)))

NODEPTR mpfr_node_atantwo(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_atan2(node_mpfrval(fp), node_mpfrval(e1),node_mpfrval(e2), MPFRRnd);
  return (fp);
}

#define int_exp(X,Y) int_node (power (node_ival(X), node_ival(Y)))
#define float_flpower(X,Y) float_node(pow(node_flval(X),node_flval(Y)))

NODEPTR mpfr_node_pow(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_pow(node_mpfrval(fp), node_mpfrval(e1),node_mpfrval(e2), MPFRRnd);
  return (fp);
}

NODEPTR mpz_node_exp_ui(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpz_pow_ui(node_mpzval(fp), node_mpzval(e1), mpz_get_ui (node_mpzval (e2)));
  return (fp);
}

NODEPTR mpx_node_exp_ui(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_pow_ui(node_mpxval(fp), node_mpxval(e1), mpz_get_ui (node_mpxval (e2)));
  return (fp);
}

NODEPTR mpx_node_mpxand(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_and(node_mpxval(fp), node_mpxval(e1),node_mpxval(e2));
  return (fp);
}

NODEPTR mpx_node_mpxior(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_ior(node_mpxval(fp), node_mpxval(e1),node_mpxval(e2));
  return (fp);
}

NODEPTR mpx_node_mpxxor(NODEPTR e1, NODEPTR e2)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_xor(node_mpxval(fp), node_mpxval(e1),node_mpxval(e2));
  return (fp);
}


/*  calculate value */
/* e1 and e2, e1 or e2 */
/* e1 < e2, e1 <= e2, e1 >= e2, e1 > e2 */ 
/* e1 + e2, e1 - e2, e1 * e2, e1 div e2, e1 mod e2, e1 ** e2 */
NODEPTR calculate_val(unsigned op, NODEPTR e1, NODEPTR e2)
{
    unsigned type = node_type(e1);
    int64_t val;
    int64_t product;
        /*double fl_product;*/

    
    
      type_check (e2, type);
      switch (op) {
      case AND:
	switch (type) {
	case BOOL:
            return (node_bval(e1) ? e2 : e1);
	default:
	  break;
	}
	break;
      case OR:
	switch (type) {
	case BOOL:
            return (node_bval(e1) ? e1 : e2);
	default:
	  break;
	}
	break;
      case LT:
          return (compare_val (e1, e2) < 0 ? True : False);
      case LE:
          return (compare_val (e1, e2) <= 0 ? True : False);
      case GE:
          return (compare_val (e1, e2) >= 0 ? True : False);
      case GT:
          return (compare_val (e1, e2) > 0 ? True : False);
      case PLUS:
	switch (type) {
	case INT:
            assert0(add_of(val,node_ival(e1),node_ival(e2))==0,"integer overflow in addition");
            return (int_node(val));
        case FLOAT:
            return (float_plus (e1,e2));
        case MPZ:
            return (mpz_node_plus (e1,e2));
        case MPX:
            return (mpx_node_plus (e1,e2));    
        case MPFR:
             return (mpfr_node_plus (e1,e2));
	case LIST:
            return (append_list (e1, e2));
	case STRING:
            return (append_string (e1, e2));
	default:
	  break;
	}
	break;
      case MINUS:
	switch (type) {
	case INT:
            assert0(sub_of(val,node_ival(e1),node_ival(e2))==0,"integer overflow in subtraction");
	  return (int_node (val));
        case FLOAT:
            return (float_minus (e1,e2));
        case MPZ:
            return (mpz_node_minus (e1,e2));
        case MPX:
            return (mpx_node_minus (e1,e2));    
        case MPFR:
            return (mpfr_node_minus (e1,e2));
	default:
	  break;
	}
	break;
      case MULT:
	switch (type) {
	case INT:
            assert0(mult (node_ival(e1), node_ival(e2), &product) == 0,"integer overflow in multiplication" );
            return (int_node(product));
        case FLOAT:
            return (float_mult (e1,e2));
        case MPZ:
          return (mpz_node_mult (e1,e2));
        case MPX:
          return (mpx_node_mult (e1,e2));  
        case MPFR:
          return (mpfr_node_mult (e1,e2));
	default:
	  break;
	}
	break;  
      case DIV:
	switch (type) {
	case INT:
            assert0(node_ival(e2) !=0, "division by zero is undefined\n"); 
	  return (int_div (e1, e2));
        case FLOAT:
          /*assert0(node_flval(e2) !=0, "division by zero is undefined\n");*/
            return (float_div (e1,e2));
        case MPZ:
          assert0(mpz_sgn (node_mpzval(e2)) != 0, "division by zero is undefined\n");
          return (mpz_node_div (e1,e2));
        case MPX:
          assert0(mpz_sgn (node_mpxval(e2)) != 0, "division by zero is undefined\n");
          return (mpx_node_div (e1,e2));  
        case MPFR:
          /*assert0(mpfr_sgn (node_mpfrval(e2)) != 0, "division by zero is undefined\n");*/
          return (mpfr_node_div (e1,e2));  
	default:
	  break;
	}
	break;
      case MOD:
	switch (type) {
	case INT:
            assert0(node_ival(e2) !=0, "division by zero is undefined\n");
	  return (int_mod (e1, e2));
        case FLOAT:
          /*assert0(node_flval(e2) !=0, "division by zero is undefined\n");*/
	  return (float_mod (e1, e2));
        case MPZ:
            assert0(mpz_sgn (node_mpzval(e2)) != 0, "division by zero is undefined\n");
          return (mpz_node_mod (e1,e2));
        case MPX:
            assert0(mpz_sgn (node_mpxval(e2)) != 0, "division by zero is undefined\n");
          return (mpx_node_mod (e1,e2));  
        case MPFR:
          /*assert0(mpfr_sgn (node_mpfrval(e2)) != 0, "division by zero is undefined\n");*/
          return (mpfr_node_mod (e1,e2)); 
	default:
	  break;
	}
	break;
      case ATANTWO:
	switch (type) {
        case FLOAT:
	  return (float_atantwo (e1, e2));
        case MPFR:
          return (mpfr_node_atantwo (e1,e2)); 
	default:
	  break;
	}
	break;  
      case POWER:
	switch (type) {
	case INT:
            assert0(node_ival(e2) >=0, "power to should be >=0\n");
            product=power(node_ival(e1), node_ival(e2));
	  return (int_node (product));
        case FLOAT:
          /*assert0(node_flval(e1)!=0 || node_flval(e2) >0, "domain error power to\n");
            assert0(node_flval(e1)>=0, "domain error power to\n");*/
            return (float_flpower(e1, e2));
        case MPZ:
          assert0(mpz_sgn (node_mpzval(e2))>=0, "domain error power to\n");
          assert0(mpz_sgn(node_mpzval(e1))>=0, "domain error power to\n");
          return (mpz_node_exp_ui (e1,e2));
        case MPX:
          assert0(mpz_sgn (node_mpxval(e2))>=0, "domain error power to\n");
          assert0(mpz_sgn(node_mpxval(e1))>=0, "domain error power to\n");
          return (mpx_node_exp_ui (e1,e2));  
        case MPFR:
          /*assert0(mpfr_sgn (node_mpfrval(e2))>=0, "domain error power to\n");
            assert0(mpfr_sgn(node_mpfrval(e1))>=0, "domain error power to\n");*/
          return (mpfr_node_pow (e1,e2));  
	default:
	  break;
	}
	break;
      case MPXAND:
	switch (type) {
          case MPX:
            return (mpx_node_mpxand(e1,e2));
          default:
            break;
        }
        break;
      case MPXIOR:
	switch (type) {
          case MPX:
            return (mpx_node_mpxior(e1,e2));
          default:
            break;
        }
        break;
      case MPXXOR:
	switch (type) {
          case MPX:
            return (mpx_node_mpxxor(e1,e2));
          default:
            break;
        }
        break;  
      default:
	break;
      }
      error_break2 ("\"%s\" is not a %s operator",
		    Keyword[op], Keyword[type]);
      return (False);
}


/* Binary expressions: AND, OR, PLUS, MINUS, ... */
/* boolean expressions are speeded up otherwise use calculate_val*/
void reduce_bin_op(NODEPTR e)
{
  /*NODEPTR calculate_val ();*/
    NODEPTR e1, e2;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        switch (node_type(e)) {
            case AND:
                e2 = (node_bval(e1) == F) ? e1 : reduce_val (node_lnk_2(e));
                break;
            case OR:
                e2 = (node_bval(e1) == T) ? e1 : reduce_val (node_lnk_2(e));
                break;
            default:
                e2 = reduce_val (node_lnk_2(e));
        }
    }
    if (Ready)
        result (calculate_val (node_type(e), e1, e2));
}

#define int_uminus(X) int_node (- node_ival(X))
#define int_not(X) (node_ival(X) == 0 ? One : Zero)
#define float_uminus(X) float_node (- node_flval(X))

NODEPTR mpz_node_uminus(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpz_neg(node_mpzval(fp), node_mpzval(e1));
  return (fp);
}

NODEPTR mpx_node_uminus(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_neg(node_mpxval(fp), node_mpxval(e1));
  return (fp);
}

NODEPTR mpfr_node_uminus(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_neg(node_mpfrval(fp), node_mpfrval(e1), MPFRRnd);
  return (fp);
}

NODEPTR mpz_node_ceil(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpfr_t temp;
  mpfr_init (temp);
  mpfr_ceil(temp, node_mpfrval(e1));
  mpfr_get_z(node_mpzval(fp), temp, MPFRRnd);
  mpfr_clear (temp);
  return (fp);
}

NODEPTR mpx_node_ceil(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpfr_t temp;
  mpfr_init (temp);
  mpfr_ceil(temp, node_mpfrval(e1));
  mpfr_get_z(node_mpxval(fp), temp, MPFRRnd);
  mpfr_clear (temp);
  return (fp);
}

NODEPTR mpz_node_floor(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpfr_t temp;
  mpfr_init (temp);
  mpfr_floor(temp, node_mpfrval(e1));
  mpfr_get_z(node_mpzval(fp), temp, MPFRRnd);
  mpfr_clear (temp);
  return (fp);
}

NODEPTR mpx_node_floor(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpfr_t temp;
  mpfr_init (temp);
  mpfr_floor(temp, node_mpfrval(e1));
  mpfr_get_z(node_mpxval(fp), temp, MPFRRnd);
  mpfr_clear (temp);
  return (fp);
}

NODEPTR mpz_node_rint(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));
  mpfr_t temp;
  mpfr_init (temp);
  mpfr_rint(temp, node_mpfrval(e1), MPFRRnd);
  mpfr_get_z(node_mpzval(fp), temp, MPFRRnd);
  mpfr_clear (temp);
  return (fp);
}

NODEPTR mpx_node_rint(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpfr_t temp;
  mpfr_init (temp);
  mpfr_rint(temp, node_mpfrval(e1), MPFRRnd);
  mpfr_get_z(node_mpxval(fp), temp, MPFRRnd);
  mpfr_clear (temp);
  return (fp);
}



NODEPTR mpfr_node_sqrt(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_sqrt (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_expo(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_exp (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_log(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_log (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_log10(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_log10 (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_sin(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_sin (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_cos(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_cos (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_tan(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_tan (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_asin(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_asin (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_acos(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_acos (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_atan(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_atan (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_sinh(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_sinh (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_cosh(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_cosh (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_tanh(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_tanh (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpfr_node_fabs(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));
  mpfr_abs (node_mpfrval(fp), node_mpfrval(e1),  MPFRRnd);
  
  return (fp);
}

NODEPTR mpz_node_itompz(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));

  mpz_set_si ( node_mpzval(fp), node_ival(e1));

  return (fp);
}

NODEPTR mpx_node_itompx(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));

  mpz_set_si ( node_mpxval(fp), node_ival(e1));

  return (fp);
}

NODEPTR mpfr_node_itompfr(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));

  mpfr_set_si ( node_mpfrval(fp), node_ival(e1), MPFRRnd);

  return (fp);
}

NODEPTR mpfr_node_mpztompfr(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));

  mpfr_set_z ( node_mpfrval(fp), node_mpzval(e1), MPFRRnd);

  return (fp);
}

NODEPTR mpx_node_mpztompx(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));

  size_t size;

  size = mpz_sizeinbase(node_mpzval(e1), 16)+2;

  char t[size];
  
  mpz_get_str (t, 16, node_mpzval(e1));

  
  
  mpz_set_str ( node_mpxval(fp), t, 16);

  return (fp);
}

NODEPTR mpz_node_mpxtompz(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPZ;
  mpz_init(node_mpzval(fp));

  size_t size;

  size = mpz_sizeinbase(node_mpxval(e1), 10)+2;

  char t[size];
  
  mpz_get_str (t, 10, node_mpxval(e1));

  mpz_set_str ( node_mpzval(fp), t, 10);

  return (fp);
}


NODEPTR mpfr_node_mpxtompfr(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));

  mpfr_set_z ( node_mpfrval(fp), node_mpxval(e1), MPFRRnd);

  return (fp);
}

NODEPTR mpfr_node_ftompfr(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPFR;
  mpfr_init(node_mpfrval(fp));

  mpfr_set_d ( node_mpfrval(fp), node_flval(e1), MPFRRnd);

  return (fp);
}

NODEPTR mpx_node_mpxcom(NODEPTR e1)
{
  register NODEPTR fp = claim_node();
  node_type(fp) = MPX;
  mpz_init(node_mpxval(fp));
  mpz_com(node_mpxval(fp), node_mpxval(e1));
  return (fp);
}

NODEPTR string_node_itostr(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = STRING;

 char    s[MAXINTSIZE];
 
 snprintf(s,MAXINTSIZE,"%" PRId64 "",node_ival(e1));
 /*printf("string = %s\n",s);*/
 node_strval(fp) = strsave(s);
 
 return (fp);
}

NODEPTR string_node_ftostr(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = STRING;

 char *s2,*fmt,*p,*p1,*p2;
 char* out=NULL;
 int ret;
    
 s2 = strsave(titoa(Precision));
 p = strapnd(s2,"e");
 fmt = strapnd("%.",p);
    
 ret = asprintf(&out,fmt,node_flval(e1));

 if (ret < 0) {
   error_break0("cannot convert float to string");
 }
 p1 = strapnd("$",out);
 p2 = strapnd(p1,"$");
   
 free_mem(fmt);
 free_mem(p);
 free_mem(s2);
 
 node_strval(fp) = strsave(p2);

 free_mem(p1);
 free_mem(p2);
 free_mem(out);
 
 return (fp);
}


NODEPTR string_node_mpztostr(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = STRING;

 size_t size;

 size = mpz_sizeinbase(node_mpzval(e1), 10)+2;
  
 char t[size], *p1;

 mpz_get_str(t, 10, node_mpzval(e1));

 if (t[0] == '-') {
   p1 = strapnd("-0z",substr(t,1,strlen(t)-1));
 } else {
   p1 = strapnd("0z",t);
 }

 node_strval(fp) = strsave(p1);
 free_mem(p1);

 return (fp);
}

NODEPTR string_node_mpxtostr(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = STRING;

 size_t size;

 size = mpz_sizeinbase(node_mpxval(e1), 16)+2;
  
 char t[size], *p1;

 mpz_get_str(t, 16, node_mpxval(e1));

 if (t[0] == '-') {
   p1 = strapnd("-0x",substr(t,1,strlen(t)-1));
 } else {
   p1 = strapnd("0x",t);
 }

 node_strval(fp) = strsave(p1);
 free_mem(p1);

 return (fp);
}

NODEPTR string_node_mpfrtostr(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = STRING;

 char *s2,*fmt,*p,*p1,*p2;
 char* out= NULL;
 int ret;

 s2 = strsave(titoa(MPFRPrecision));
 p = strapnd(s2,"Re");
 fmt = strapnd("%.",p);

 ret = mpfr_asprintf(&out,fmt,node_mpfrval(e1));

 if (ret < 0) {
   error_break0("cannot convert big float to string");
 }

 p1 = strapnd("$$",out);
 p2 = strapnd(p1,"$$");

 node_strval(fp) = strsave(p2);

 mpfr_free_str(out);
 free_mem(fmt);
 free_mem(p);
 free_mem(s2);
 free_mem(p1);
 free_mem(p2);
 
 return (fp);
}

NODEPTR node_int_strtoi(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = INT;

 int64_t cint;
 char *temp;

 errno = 0;
 
 cint = strtoll(node_strval(e1),&temp,10);

 if (temp ==  node_strval(e1) || *temp != '\0')
   syntax_error(ERI);
 if ((cint == LLONG_MIN || cint == LLONG_MAX) && errno == ERANGE)
      syntax_error(ERA);
 
 node_ival(fp) = cint;

 return (fp);
}

NODEPTR node_float_strtofl(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = FLOAT;

 double cfl;
 
 /*printf("length = %lu\n", strlen(node_strval(e1)));*/
 
 if ( strlen(node_strval(e1)) <=2 ) {
   syntax_error(EIF);
   cfl = 0;
 } else {
   /*printf("first: %c\n", (node_strval(e1))[0]);*/
   /*printf("last: %c\n", (node_strval(e1))[strlen(node_strval(e1))-1]);*/
 
   if ( strlen(node_strval(e1)) > 2 &&
        ( (node_strval(e1))[0] != '$' ||
          (node_strval(e1))[strlen(node_strval(e1))-1] != '$') ) {
     syntax_error(EIF);  
     cfl = 0;
   } else {
     char *temp, *s;
     
     errno = 0;
     s = substr(node_strval(e1), 1, strlen(node_strval(e1))-2);
     /*printf("substr: %s\n", s);*/
 
     cfl = strtod(s,&temp);

     if (temp == s || *temp != '\0') {
       syntax_error(EIF);
       cfl = 0;
     } 

     free_mem(s);
   }
 }
 
 node_flval(fp) = cfl;
 
 return (fp);
}

NODEPTR node_mpz_strtompz(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = MPZ;

 char *s,*t;

 /*printf("first %c\n", (node_strval(e1))[0]);*/

 if ((node_strval(e1))[0] == '-') {
   s = substr(node_strval(e1), 1, strlen(node_strval(e1))-1);
   t = strapnd("-",substr(s,2,strlen(s)-1));
 } else {
   s = strsave(node_strval(e1));
   t = substr(s,2,strlen(s)-1);
 }

 /*printf("length = %lu\n", strlen(s));*/

 if (strlen(s) <= 2) {
   syntax_error(EIMPZ);
   mpz_init_set_str (node_mpzval(fp), "0", 10);
 } else {
   
   /*printf("substr 1 %s\n", substr(s,0,1));*/
   /*printf("comp %d\n", strcmp(substr(s,0,1), "0z"));*/

   if (strcmp(substr(s,0,1), "0z") !=0) {
     syntax_error(EIMPZ);
     mpz_init_set_str (node_mpzval(fp), "0", 10);
   } else {
     /*printf("substr 2 %s\n", substr(s,2,strlen(s)-1));*/
     /*printf("substr 3 %s\n", t);*/

     if (mpz_init_set_str (node_mpzval(fp),  t, 10) == -1) {
       syntax_error(EIMPZ);
       mpz_set_str (node_mpzval(fp), "0", 10);
     }
   }  
 }
 
 free_mem(s);
 free_mem(t);
 
 return (fp);
}

NODEPTR node_mpx_strtompx(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = MPX;

 char *s,*t;

 /*printf("first %c\n", (node_strval(e1))[0]);*/

 if ((node_strval(e1))[0] == '-') {
   s = substr(node_strval(e1), 1, strlen(node_strval(e1))-1);
   t = strapnd("-",substr(s,2,strlen(s)-1));
 } else {
   s = strsave(node_strval(e1));
   t = substr(s,2,strlen(s)-1);
 }

 /*printf("length = %lu\n", strlen(s));*/

 if (strlen(s) <=2) {
   syntax_error(EIMPX);
   mpz_init_set_str (node_mpxval(fp), "0", 16);
 } else {
   /*printf("substr 1 %s\n", substr(s,0,1));*/
   /*printf("comp %d\n", strcmp(substr(s,0,1), "0x"));*/

   if (strcmp(substr(s,0,1), "0x") !=0) {
     syntax_error(EIMPX);
     mpz_init_set_str (node_mpxval(fp), "0", 16);
   } else {
     /*printf("substr 2 %s\n", substr(s,2,strlen(s)));*/
     /*printf("substr 3 %s\n", t);*/

     if (mpz_init_set_str (node_mpxval(fp),  t, 16) == -1) {
       syntax_error(EIMPX);
       mpz_set_str (node_mpxval(fp), "0", 16);
     }
   }
 }
 
 free_mem(s);
 free_mem(t);
   
 return (fp);
}

NODEPTR node_mpfr_strtompfr(NODEPTR e1)
{
 register NODEPTR fp = claim_node();
 node_type(fp) = MPFR;

 char  *s;

 /*printf("length = %lu\n", strlen(node_strval(e1)));*/
 
 if (strlen(node_strval(e1)) <= 4) {
   syntax_error(EIMPFR);
   mpfr_init_set_str (node_mpfrval(fp), "0", 10, MPFRRnd);
 } else {
   /*printf("first two: %s\n", substr(node_strval(e1), 0, 1));*/
   /*printf("last  two: %s\n", substr(node_strval(e1), strlen(node_strval(e1))-2, strlen(node_strval(e1))-1));*/
 
   if (strcmp(substr(node_strval(e1), 0, 1), "$$") !=0 ||
       strcmp( substr(node_strval(e1), strlen(node_strval(e1))-2, strlen(node_strval(e1))-1), "$$") != 0) {
     syntax_error(EIMPFR);
     mpfr_init_set_str (node_mpfrval(fp), "0", 10, MPFRRnd);
   } else {
     
     s = substr(node_strval(e1), 2, strlen(node_strval(e1))-3);
     /*printf("substr: %s\n", s);*/
 
     if (mpfr_init_set_str (node_mpfrval(fp), s, 10, MPFRRnd) == -1) {
       syntax_error(EIMPFR);
       mpfr_set_str(node_mpfrval(fp), "0", 10, MPFRRnd);
     }        
 
     free_mem(s);
   }
 }
 
 return (fp);
}

NODEPTR strtoval(NODEPTR e1)
{
  char *s;
  NODEPTR val, tval;
  
  s = strapnd(node_strval(e1),".");

  tval = parse_string(s);
  if (yyerrorcount == 0) {
    val = reduce_val (tval);
  } else {
    error_break0("error in input");
  }
  free_mem(s);
  
  return (val);
}


/* unary operators: NOT, UPLUS, UMINUS, ... */
/* not e1, +e1, -e1 */
void reduce_u_op(NODEPTR e)
{
    NODEPTR e1;

    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        switch (node_type(e1)) {
            case STRING:
                switch (node_type(e)) {
                    case STRTOVAL:
                      result (strtoval(e1));
                      break;                      
                    case STRTOI:
                      result (node_int_strtoi(e1));
                      break;
                    case STRTOF:
                      result (node_float_strtofl(e1));
                      break;
                    case STRTOMPZ:
                      result (node_mpz_strtompz(e1));
                      break;
                    case STRTOMPX:
                      result (node_mpx_strtompx(e1));
                      break;
                    case STRTOMPFR:
                      result (node_mpfr_strtompfr(e1));
                      break; 
                    default:
                      error_break1 ("\"%s\" is not a string operator",
                                print_name (e));
                }
                break;
            case BOOL:
                switch (node_type(e)) {
                    case NOT:
                        result (node_bval(e1) ? False : True);
                        break;
                    default:
                        error_break1 ("\"%s\" is not a Boolean operator",
                                print_name (e));
                }
                break;
            case INT:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (int_uminus (e1));
                        break;
                    case ITOF:
                        result (float_node((double)node_ival(e1)));
                        break;
                    case ITOSTR:
                        result (string_node_itostr(e1));
                        break;
                    case ITOMPZ:
                        result (mpz_node_itompz (e1));
                        break;
                    case ITOMPX:
                        result (mpx_node_itompx (e1));
                        break;    
                    case ITOMPFR:
                        result (mpfr_node_itompfr (e1));
                        break;    
                    default:
                        error_break1 ("\"%s\" is not an integer operator",
                                   print_name (e));
                }
                break;
             case FLOAT:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (float_uminus (e1));
                        break;
                    case CEIL:
                        result (int_node(ceil(node_flval(e1))));
                        break;
                    case FLOOR:
                        result (int_node(floor(node_flval(e1))));
                        break;
                    case RINT:
                        result (int_node(lrint(node_flval(e1))));
                        break;
                    case SQRT:
                        result (float_node(sqrt(node_flval(e1))));
                        break;
                    case EXPO:
                        result (float_node(exp(node_flval(e1))));
                        break;
                    case LOG:
                      /* assert0(node_flval(e1)>0, "log argument should be greater than zero\n");*/
                        result (float_node(log(node_flval(e1))));
                        break;
                    case LOGTEN:
                      /*assert0(node_flval(e1)>0, "log10 argument should be greater than zero\n");*/
                        result (float_node(log10(node_flval(e1))));
                        break;
                    case SIN:
                        result (float_node(sin(node_flval(e1))));
                        break;
                    case COS:
                        result (float_node(cos(node_flval(e1))));
                        break;
                    case TAN:
                        result (float_node(tan(node_flval(e1))));
                        break;
                    case ASIN:
                      /*assert0(node_flval(e1)>=-1 && node_flval(e1)<=1, "asin argument should be in range [-1,1]\n");*/
                        result (float_node(asin(node_flval(e1))));
                        break;
                    case ACOS:
                      /*assert0(node_flval(e1)>=-1 && node_flval(e1)<=1, "acos argument should be in range [-1,1]\n");*/
                        result (float_node(acos(node_flval(e1))));
                        break;
                    case ATAN:
                        result (float_node(atan(node_flval(e1))));
                        break;
                    case SINH:
                        result (float_node(sinh(node_flval(e1))));
                        break;
                    case COSH:
                        result (float_node(cosh(node_flval(e1))));
                        break;
                    case TANH:
                        result (float_node(tanh(node_flval(e1))));
                        break;
                    case FABS:
                        result (float_node(fabs(node_flval(e1))));
                        break;
                    case FTOMPFR:
                        result(mpfr_node_ftompfr(e1));
                        break;
                    case FTOSTR:
                        result (string_node_ftostr(e1));
                        break;
                    default:
                        error_break1 ("\"%s\" is not an floating point operator",
                                   print_name (e));
                }
                break;
            case MPFR:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (mpfr_node_uminus (e1));
                        break;
                    case CEIL:
                        result (mpz_node_ceil(e1));
                        break;
                    case FLOOR:
                        result (mpz_node_floor(e1));
                        break;
                    case RINT:
                        result (mpz_node_rint(e1));
                        break;    
                    case SQRT:
                        result (mpfr_node_sqrt(e1));
                        break;
                    case EXPO:
                        result (mpfr_node_expo(e1));
                        break;
                    case LOG:
                        result (mpfr_node_log(e1));
                        break;
                    case LOGTEN:
                        result (mpfr_node_log10(e1));
                        break;
                    case SIN:
                        result (mpfr_node_sin(e1));
                        break;
                    case COS:
                        result (mpfr_node_cos(e1));
                        break;
                    case TAN:
                        result (mpfr_node_tan(e1));
                        break;
                    case ASIN:
                        result (mpfr_node_asin(e1));
                        break;
                    case ACOS:
                        result (mpfr_node_acos(e1));
                        break;
                    case ATAN:
                        result (mpfr_node_atan(e1));
                        break;
                    case SINH:
                        result (mpfr_node_sinh(e1));
                        break;
                    case COSH:
                        result (mpfr_node_cosh(e1));
                        break;
                    case TANH:
                        result (mpfr_node_tanh(e1));
                        break;
                    case FABS:
                        result (mpfr_node_fabs(e1));
                        break;
                    case MPFRTOSTR:
                        result (string_node_mpfrtostr(e1));
                        break;
                    case MPFRTOF:
                      result (float_node(mpfr_get_d(node_mpfrval(e1), MPFRRnd)));
                         break;
                    default:
                        error_break1 ("\"%s\" is not an big floating point operator",
                                   print_name (e));
                }
                break;    
            case MPZ:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (mpz_node_uminus (e1));
                        break;
                    case MPZTOI:
                      assert0( mpz_fits_slong_p (node_mpzval (e1)) != 0 , "mpz integer too big to convert to integer\n");
                        result (int_node( mpz_get_si(node_mpzval (e1))));
                        break;
                    case MPZTOMPX:
                        result (mpx_node_mpztompx (e1));
                        break;
                    case MPZTOMPFR:
                        result (mpfr_node_mpztompfr (e1));
                        break;
                    case MPZTOSTR:
                        result (string_node_mpztostr(e1));
                        break;
                    default:
                        error_break1 ("\"%s\" is not a big integer operator",
                                   print_name (e));
                }
                break;
          case MPX:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (mpx_node_uminus (e1));
                        break;
                    case MPXTOI:
                      assert0( mpz_fits_slong_p (node_mpxval (e1)) != 0 , "mpz hexadecimal too big to convert to integer\n");
                        result (int_node( mpz_get_si(node_mpxval (e1))));
                        break;
                    case MPXTOMPZ:
                        result (mpz_node_mpxtompz (e1));
                        break;    
                    case MPXTOMPFR:
                        result (mpfr_node_mpxtompfr (e1));
                        break;
                    case MPXCOM:
                        result (mpx_node_mpxcom (e1));
                        break;    
                    case MPXTOSTR:
                        result (string_node_mpxtostr(e1));
                        break;    
                    default:
                        error_break1 ("\"%s\" is not a big hexadecimal operator",
                                   print_name (e));
                }
                break;
                    /*case STRING:
                switch (node_type(e)) {
                        case SIZE:
                        result (int_node ((int64_t)node_lnk_2(e)));
                        break;
                    default:
                        error_break1 ("\"%s\" is not a string operator",
                                print_name (e));
                }
                break;*/
            default:
                error_break2 ("\"%s\" is not a %s operator",
                        print_name (e), print_name (e1));
        }
    }
}


/* state_number  */
/* Reduces to current state number */
void reduce_STATENUMBER(void)
{
  if ( StateNrValTime != Clock || StateNrVal == -1 ) {
    StateNrValTime = Clock;
  };
  Ready = T;
  result (int_node (StateNumber));
}


/* random  */
/* Reduces to random integer >=0  */
void reduce_RANDOM(void)
{
  if ( RandomValTime != Clock || RandomVal == -1 ) {
          /*RandomVal = i4_uniform_0i (&TRandSeed);*/
          /*RandomVal = genrand64_int63();*/
      RandomVal = xorshift128plus();
      RandomValTime = Clock;
  };
  Ready = T;
  result (int_node (RandomVal));
}

/* zrandom  */
/* Reduces to random big integer >=0  and <  zrandommax */
void reduce_ZRANDOM(void)
{
  if ( ZRandomValTime != Clock ) {
          /*RandomVal = i4_uniform_0i (&TRandSeed);*/
          /*RandomVal = genrand64_int63();*/
    /*RandomVal = xorshift128plus();*/
      mpz_urandomm (ZRandomVal, Randstate, ZRandomMax);
      ZRandomValTime = Clock;
  };
  Ready = T;
  result (mpz_node (ZRandomVal));
}

/* frrandom  */
/* Reduces to random big float >=0  and <= 1 */
void reduce_FRANDOM(void)
{
  if ( FRandomValTime != Clock ) {
          /*RandomVal = i4_uniform_0i (&TRandSeed);*/
          /*RandomVal = genrand64_int63();*/
    /*RandomVal = xorshift128plus();*/
      mpfr_urandom (FRandomVal, Randstate, MPFRRnd);
      FRandomValTime = Clock;
  };
  Ready = T;
  result (mpfr_node (FRandomVal));
}




/* Random  */
/* Reduces to random integer >=0  */
/* This will  allow for more than 1 random number
   in the same state */
void reduce_RRANDOM(void)
{
  if ( RRandomValTime != Clock || RRandomVal == -1 ) {
    RRandomValTime = Clock;
  };
  Ready = T;
      /* result (int_node (i4_uniform_0i (&TRRandSeed)));*/
      /*result (int_node (genrand64_int63()));*/
  result (int_node(xorshift128plus()));
  
}

/* zRandom  */
/* Reduces to random big integer >=0  and < zrandommax */
/* This will  allow for more than 1 random number
   in the same state */
void reduce_ZRRANDOM(void)
{
  if ( ZRRandomValTime != Clock  ) {
    ZRRandomValTime = Clock;
  };
  Ready = T;
      /* result (int_node (i4_uniform_0i (&TRRandSeed)));*/
      /*result (int_node (genrand64_int63()));*/
  /*result (int_node(xorshift128plus()));*/
  mpz_urandomm (ZRRandomVal, Randstate, ZRandomMax);
  result (mpz_node (ZRRandomVal));
}

/* frRandom  */
/* Reduces to random big float >=0  and <= 1 */
/* This will  allow for more than 1 random number
   in the same state */
void reduce_FRRANDOM(void)
{
  if ( FRRandomValTime != Clock ) {
          /*RandomVal = i4_uniform_0i (&TRandSeed);*/
          /*RandomVal = genrand64_int63();*/
    /*RandomVal = xorshift128plus();*/
      FRandomValTime = Clock;
  };
  Ready = T;
  mpfr_urandom (FRRandomVal, Randstate, MPFRRnd);
  result (mpfr_node (FRRandomVal));
}


/* frandom  */
/* Reduces to random in [0.0,1.0) */
void reduce_SRANDOM(void)
{
  if ( SRandomValTime != Clock || SRandomVal == -1 ) {
          /*SRandomVal = r4_uniform_01 (&TSRandSeed);*/
          /*SRandomVal = genrand64_real2();*/
      SRandomVal = float_xorshift128plus();
      SRandomValTime = Clock;
  };
  Ready = T;
  result (float_node (SRandomVal));
}

/* fRandom  */
/* Reduces to random in [0.0,1.0) */
/* This will  allow for more than 1 frandom number
   in the same state */
void reduce_SRRANDOM(void)
{
  if ( SRRandomValTime != Clock || SRRandomVal == -1 ) {
    SRRandomValTime = Clock;
  };
  Ready = T;
      /* result (float_node (r4_uniform_01 (&TSRRandSeed)));*/
      /*result (float_node ( genrand64_real2() ));*/
  result (float_node(float_xorshift128plus()));
  
}

/********************\
*                    *
*  string operators  *
*                    *
\********************/

/* ascii(e1) */
void reduce_ASCII(NODEPTR e)
{
    char   *strr;
    NODEPTR e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (e1, STRING);
        assert1 (strlen (strr = node_strval(e1)) == 1,
                "cannot get ascii code of string \"%s\" (length not 1)", strr);
        result (int_node (*strr));
    }
}



/**********************\
*                      *
*  pointer operations  *
*                      *
\**********************/
/* ref(loc) */
void reduce_REFER(NODEPTR e)
{
    NODEPTR loc = reduce_ref (node_lnk_1(e));
    if (Ready) {
        assert0 (loc != TempLoc, "cannot reference a value");
        result (loc_node (loc, is_next(loc))); 
    }
}
/* deref p */
void reduce_DEREF(NODEPTR e)
{
    NODEPTR p = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (p, LOC); 
        ref_result (node_loc_tonext(p) ? next_loc (node_loc_ptr(p)) : curr_loc (node_loc_ptr(p)));
    }
}

