/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-int.c 284 2017-10-08 08:00:31Z cau $
 */



#include "system.h"




/****************************************\
* 					*
* interval-dependent control statements *
* 	                                *
\****************************************/


/* SEMI w1 w2 */
void reduce_SEMI(NODEPTR e)
{
    NODEPTR e1 = e;
    NODEPTR empty_loc = local_mem (STATE);
    for (e = NULL; node_type(e1) == SEMI; e1 = node_lnk_1(e1))
        e = node3 (iSEMI, empty_loc, node_lnk_2(e1), e); 
    Retry = T;
    result (node3 (iSEMI, empty_loc, e1, e)); 
}


/* iSEMI empty_loc w next */
void reduce_iSEMI(NODEPTR e)
{
    NODEPTR empty_loc;

    if (DebugLevel>=5)
        printf("push_empty\n");
    empty_loc = node_blk_base(node_lnk_1(e));
    push_empty (empty_loc);
    node_lnk_2(e) = reduce_pgm (node_lnk_2(e));
    if (DebugLevel>=5)
        printf("pop_empty\n");
    pop_empty;
    if (mem_defined1 (empty_loc = curr_loc (empty_loc))) {
        if (get_mem (empty_loc) == False) {
            put_empty (False);
        }
        else {
            if (Ready) {
                if ((e = node_lnk_3(e)) == NULL) {
                    put_empty (True);
                    result (True);
                }
                else {
                    Retry = T;
                    if (!RedundancyFlag && node_lnk_3(e) == NULL) {
                        result (node_lnk_2(e));
                    }
                    else {
                        erase_mem (empty_loc);
                        result (e);
                    }
                }
            }
            else {
                if (HopChop && !PassProgressFlag) {
                    if (!RedundancyFlag && node_lnk_3(e) == NULL ) {
                        if (DebugLevel>=5) {
                            printf("suffix chop is null \n");
                        }
                    }
                    else {
                        erase_mem(empty_loc);
                        Retry=T;
                        result(node2(AND, node_lnk_3(e), node2(SEMI,node2(AND,node_lnk_2(e),Empty),True)));
                    }
                }
            }
        }
        
    }
}


/* WHILE b s */
void reduce_WHILE(NODEPTR e)
{
    NODEPTR b = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (b, BOOL);
        result (node_bval(b) ? node2 (SEMI, node_lnk_2(e), e) : Empty);
        Retry = T;
    }
}

/* REPEAT s b */
void reduce_REPEAT(NODEPTR e)
{
    Retry = T;
    result (node2 (SEMI, node_lnk_1(e),
    	node2 (WHILE, node1 (NOT, node_lnk_2(e)), node_lnk_1(e))));
}

/* FORTIMES e1 e2 */
void reduce_FORTIMES(NODEPTR e)
{
    NODEPTR e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (e1, INT);
        Retry = T;
        result (node2 (WHILE,
                       node1(iFORTIMES, int_node (node_ival(e1))), node_lnk_2(e)));
    }
}

/* iFORTIMES e1 */
void reduce_iFORTIMES(NODEPTR e)
{
    NODEPTR e1 = node_lnk_1(e);
    if (node_ival(e1) == 0)
        result (False);
    else {
        node_ival(e1) -= 1;
        result (True);
    }
}

/* iFORTIMES i e2
reduce_iFORTIMES(e)
NODEPTR e;
{
    if (node_lnk_1(e) -> ival == 0)
    result (Empty);
    else {
    node_lnk_1(e) -> ival -= 1;
    result (node2 (SEMI, node_lnk_2(e), e));
    }
    Retry = T;
}
*/


/* FOR id n pgm */
void reduce_FOR(NODEPTR e)
{
    NODEPTR env = node_lnk_1(e);
    NODEPTR blk;
    NODEPTR e2 = reduce_val (node_lnk_2(e));
    if (Ready) {
        type_check (e2, INT);
        static_check (node_env_idlist(env) );
        blk = local_mem (node_type(node_env_idlist(env)));
        result (node2 (iEXISTS, bnd_node (env, blk, NULL), 
                       node2 (WHILE,
                              node3 (iFOR, int_node (0), e2, blk),
                              node_lnk_3(e))));                 
        Retry = T;
    }
}

/* iFOR i n loc */
void reduce_iFOR(NODEPTR e)
{
    NODEPTR loc = node_blk_base(node_lnk_3(e));
    int       i;
    if ((i = node_ival(node_lnk_1(e)) ++) >= node_ival(node_lnk_2(e)) )
        result (False);
    else {
        put_mem1 (loc, int_node (i));
        result (True);
    }
}

/* FORIN id list expn */
void reduce_FORIN(NODEPTR e)
{
    NODEPTR env = node_lnk_1(e);
    NODEPTR blk;
    NODEPTR list = reduce_val (node_lnk_2(e));
    if (Ready) {
        static_check (node_env_idlist(env) );
        blk = local_mem (node_type(node_env_idlist(env)));
        Retry = T;
        result (node2 (iEXISTS, bnd_node (env, blk, NULL), 
                       node2 (WHILE,
                              node3 (iFORIN, int_node (0), list, blk),
                              node_lnk_3(e))));                 
    }
}

/* iFORIN i list base_loc */
void reduce_iFORIN(NODEPTR e)
{
    NODEPTR loc = node_blk_base(node_lnk_3(e));
    NODEPTR list = node_lnk_2(e);
    unsigned int i;
    switch (node_type(list)) {
      case LIST:
        if ((i = (node_ival(node_lnk_1(e)))++) >= (unsigned int) get_int (list_size (list)))
            result (False);
        else {
            make_indirect_mem (loc, list_base (list) + offset2 (i,node_mem_history(list_base (list))));
                /*printf("reduce_iFORIN: base_loc=%p\n",list_base (list));*/
            result (True);
        }
        break;
      case STRING:      
        if ((i = (node_ival(node_lnk_1(e)) )++) >= (unsigned int) string_size (list))
            result (False);
        else {        
            put_mem1 (curr_loc (loc), substring (list, (int)i, (int)i));
            result (True);
        }
        break;
      default:
        error_break1 ("cannot index expression of type: %s",
            print_name (list));
    }
}

/* CHOPSTAR e */
void reduce_CHOPSTAR(NODEPTR e)
{
  if ((Ready = empty_defined ())) {
    if (get_empty () == True)
      result (True);
    else {
      Retry = T;
      result (node2 (SEMI, node2 (AND, node_lnk_1(e), More), e));
    }
  }
}

/* KEEP s */
/* BCM 15Aug97 */
void reduce_KEEP(NODEPTR e)
{
  if ((Ready = empty_defined ())) {
    if (get_empty () == True)
      result (True);
    else {
      Retry = T;
      result (node2 (SEMI, node2 (AND, Skip, node_lnk_1(e)), e));
    }
  }
}

/************************\
*               	*
*  Projection operators	*
* 			*
\************************/

/* PROJ e1 e2 */
void reduce_PROJ(NODEPTR e)
{
/*    assert0 (!AssignAheadFlag,*/
/*        "projection requires \"assign_ahead\" to be false");*/
    Retry = T;
    result (node3 (iPROJ, False, node_lnk_1(e), node_lnk_2(e)));
}

/* iPROJ ready e1 e2 */
void reduce_iPROJ(NODEPTR e)
{
    if (!(Ready = node_bval(node_lnk_1(e)) )) {
        push_assign_ahead;
        node_lnk_3(e) = reduce_stmt (node_lnk_3(e));
        node_lnk_1(e) = bool_node (Ready);
        pop_assign_ahead;
    }
    if (Ready) {
        if ((Ready = empty_defined ())) {
            if (get_empty () == True) {
                result (True);
            }
            else {
                Retry = T;
		node_lnk_1(e) = False;
                result (node2 (SEMI, node_lnk_2(e), e));
	    }
        }
    }
}


